/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-cert.c,v 1.5 2009/08/09 16:11:41 ca Exp $")

#include "sm/error.h"
#include "sm/sysexits.h"
#include "sm/io.h"
#include "sm/tls.h"
#include "sm/test.h"

static int Verbose = 0;

#if MTA_USE_TLS
/* based on openssl/apps/apps.c, simplified for testing purposes */
static X509 *
load_cert(BIO *err, const char *file)
{
	X509 *x;
	BIO *cert;

	x = NULL;
	cert = BIO_new(BIO_s_file());
	if (cert == NULL) {
		ERR_print_errors(err);
		goto end;
	}

	if (BIO_read_filename(cert,file) <= 0) {
		BIO_printf(err, "Error opening %s\n", file);
		ERR_print_errors(err);
		goto end;
	}

	x = PEM_read_bio_X509_AUX(cert,NULL, (pem_password_cb *)NULL, NULL);
end:
	if (x == NULL) {
		BIO_printf(err,"unable to load certificate\n");
		ERR_print_errors(err);
	}
	if (cert != NULL)
		BIO_free(cert);
	return x;
}

#define MAXN 	256

static void
t_cert_x509(const char *infile, const char *cn_subject, const char *cn_issuer, const char *dn_subject)
{
	sm_ret_T ret;
	X509_STORE *ctx;
	X509 *x509_cert=NULL;
	BIO *bio_err = NULL;
	X509_NAME *subj, *issuer;
	sm_str_P str;

	ctx = X509_STORE_new();
	SM_TEST_E(ctx != NULL);
	str = sm_str_new(NULL, MAXN, MAXN * 2);
	SM_TEST_E(str != NULL);

	ERR_load_crypto_strings();
	bio_err=BIO_new_fp(stderr,BIO_NOCLOSE);
	SM_TEST_E(bio_err != NULL);

	x509_cert = load_cert(bio_err, infile);
	SM_TEST_E(x509_cert != NULL);
	subj = X509_get_subject_name(x509_cert);
	SM_TEST_E(subj != NULL);
	issuer = X509_get_issuer_name(x509_cert);
	SM_TEST_E(issuer != NULL);

	sm_str_clr(str);
	ret = sm_tls_get_cn(subj, NID_commonName, str);
	SM_TEST(sm_is_success(ret));
	SM_TEST(strcmp(sm_str_getdata(str), cn_subject) == 0);
	if (Verbose > 1)
		sm_io_fprintf(smioerr, "cn_subject=%S, expected=%s\n", str, cn_subject);

	sm_str_clr(str);
	ret = sm_tls_get_cn(issuer, NID_commonName, str);
	SM_TEST(sm_is_success(ret));
	SM_TEST(strcmp(sm_str_getdata(str), cn_issuer) == 0);
	if (Verbose > 1)
		sm_io_fprintf(smioerr, "cn_issuer=%S, expected=%s\n", str, cn_issuer);

	if (dn_subject != NULL) {
		int r;
		char buf[SM_X509_MAXNAME];

		X509_NAME_oneline(subj, buf, sizeof buf);
		TLS_SET_XSTR(buf, str);
		SM_TEST(strcmp(sm_str_getdata(str), dn_subject) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "dn_subject=%S, expected=%s\n",
				str, dn_subject);
	}

  error:
	return;
}
#endif /* MTA_USE_TLS */

extern int optind;

int
main(int argc, char *argv[])
{
	int c;
	bool interactive;
	char *infile, *cn_s, *cn_i, *dn_s;

	infile = "cert.pem";
	cn_s = "host.example.org";
	cn_i = "CA";
	dn_s = NULL;
	interactive = false;
	while ((c = getopt(argc, argv, "d:i:s:V")) != -1)
	{
		switch (c)
		{
		  case 'd':
			dn_s = strdup(optarg);
			break;
		  case 'i':
			cn_i = strdup(optarg);
			break;
		  case 's':
			cn_s = strdup(optarg);
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			return EX_USAGE;
		}
	}
	sm_test_begin(argc, argv, "test cert X509 names");
	argc -= optind;
	argv += optind;
	if (argc > 0)
		infile = argv[0];

#if MTA_USE_TLS
	t_cert_x509(infile, cn_s, cn_i, dn_s);
#endif
	return sm_test_end();
}
