/*
 * Copyright (c) 2003-2006 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-mm.c,v 1.43 2008/01/09 05:28:37 ca Exp $")

#include "sm/error.h"
#include "sm/sysexits.h"
#include "sm/ctype.h"
#include "sm/heap.h"
#include "sm/memops.h"
#include "sm/maps.h"
#include "sm/mapc.h"
#include "sm/map.h"
#include "sm/mapclasses.h"
#include "sm/bdb.h"
#include "sm/cdb_map.h"
#include "sm/net.h"
#include "sm/rfc2821.h"

#include "sm/io.h"

static int Verbose = 0;
static uint line = 0;

#define SEP	':'
#define SEP_WSPC	(-1)
#define MAPC_TYPE	"hash"
#define MAPC_NAME	"aliases"
#define MAPC_FILE	"aliases.db"
#define MAX_IN_LEN	(256 * 1024)
#define SM_COMM_CHAR	'#'

#define SMM_FL_LWR_KEY		0x00000001u /* change key to lower case */
#define SMM_FL_SKIP_LEAD_SP	0x00000002u /* skip leading whitespace on RHS */
#define SMM_FL_WHSP_DELIM	0x00000004u /* use whitespace as delimiter */
#define SMM_FL_EMPTY_KEY	0x00000008u /* allow empty keys */
#define SMM_FL_SKIP_COMM	0x00000010u /* skip comments */
#define SMM_FL_STRICT_SYNTAX_CHECK 0x00000020u /* strict syntax check */

#ifndef SM_ALLOW_ROOT
# define SM_ALLOW_ROOT 0
#endif

struct prefix_tab {
	const char *name;
	int (*check_fn) (const char *, uint);
};

static const struct prefix_tab *
find_prefix(struct prefix_tab *tab, const char *pfx, uint pfxlen)
{
	for (; tab->name; tab++) {
		size_t len = strlen(tab->name);
		if (len == pfxlen && memcmp(tab->name, pfx, pfxlen) == 0)
			return tab;
	}
	return NULL;
}

/* Valid addresses for from: and to: are RFC 2821 addresses
   without the angle backets (localpart@domain) as well as
   partial addresses in the form localpart and @domain... */
static int
check_email(const char *buf, uint len)
{
	uint i;
	sm_a2821_T addr;
	sm_ret_T ret;
	sm_str_T str;
	sm_rpool_P rpool;
	
	if (buf[0] == '<') {
		fprintf(stderr, "%u: %.*s: do not use angle brackets\n",
			line, len, buf);
		return 1;
	}
	if (buf[0] == '@')
		return 0;
	else if (strchr(buf, '@') == 0) {
		if (Verbose && strchr(buf, '.')) 
			fprintf(stderr, "%u: warning: suspicious localpart\n",
				line);
		return 0;
	}
	sm_str_assign(str, NULL, (uchar *)buf, len, len);
	rpool = sm_rpool_new(NULL);
	A2821_INIT_RP(&addr, rpool);
	ret = t2821_scan((sm_rdstr_P)&str, &addr, 0);
	if (!sm_is_error(ret))
		ret = t2821_parse(&addr, R2821_CORRECT&~R2821_ANGLE);
	a2821_free(&addr);
	sm_rpool_delete(rpool);
	if (sm_is_error(ret)) {
		fprintf(stderr, "%u: %.*s: invalid e-mail address\n",
			line, len, buf);
		return 1;
	}
	return 0;
}
	
static int
check_ipaddr(const char *buf, uint len)
{
	char ipbuf[16];
	ipv4_T ipv4;
		
	if (len < sizeof(ipbuf)) {
		memcpy(ipbuf, buf, len);
		ipbuf[len] = 0;

		if (!sm_is_error(sm_inet_a2ipv4(ipbuf, NULL, &ipv4)))
			return 0;
	}
		
	fprintf(stderr, "%u: %.*s: invalid IPv4 address\n", line, len, buf);
	return 1;
}
		

static struct prefix_tab access_pfx_tab[] = {
	{ "from", check_email },
	{ "to", check_email },
	{ "cltaddr", check_ipaddr },
	{ "cltname", NULL },
	{ "cltresolve", NULL },
	{ "mxbadip", check_ipaddr },
	{ "certissuer", NULL },
	{ "certsubject", NULL },
	{ "protectedrcpt", NULL },
	{ "smtps_session_conf", NULL },
	{ "smtpc_rcpt_conf", NULL },
	{ NULL }
};

#define QUICK_PFX "quick:"
#define QUICK_PFX_LEN (sizeof(QUICK_PFX) - 1)

#define ERROR_PFX "error:"
#define ERROR_PFX_LEN (sizeof(ERROR_PFX) - 1)

static int
is_access_rhs(const char *start, uint len)
{
	static char *rhs[] = { "relay", "ok", "reject", "discard",
			       "cont", NULL };
	int i;

	for (i = 0; rhs[i]; i++)
		if (strlen(rhs[i]) == len && memcmp(rhs[i], start, len) == 0)
			return 1;

	if (len > ERROR_PFX_LEN
	    && memcmp(start, ERROR_PFX, ERROR_PFX_LEN) == 0) {
		/* FIXME: Check for XYZ A.B.C.D format */
		return 1;
	}
	return 0;
}

int
check_access_syntax (sm_str_P key, sm_str_P data)
{
	uint i, len, col = 0;
	const char *start, *p;
	const struct prefix_tab *pfx;
	
	len = sm_str_getlen(key);
	start = sm_str_data(key);

	for (i = 0; i < len; i++)
		if (start[i] == ':') {
			col = i;
			break;
		}
	if (!col) {
		fprintf(stderr, "%u: %.*s: no prefix in key\n", line,
			len, start);
		return 1;
	}

	/* Check the key */
	pfx = find_prefix(access_pfx_tab, start, col);
	if (!pfx) {
		fprintf(stderr, "%u: %.*s: unknown prefix\n", line,
			col, start);
		return 1;
	}
	if (pfx->check_fn && pfx->check_fn(start + col + 1, len - col - 1))
		return 1;

	/* Check value */
	len = sm_str_getlen(data);
	start = sm_str_data(data);

	if (len > QUICK_PFX_LEN
	    && memcmp(start, QUICK_PFX, QUICK_PFX_LEN) == 0) {
		len -= QUICK_PFX_LEN;
		start += QUICK_PFX_LEN;
	}

	if (!is_access_rhs(start, len)) {
		fprintf(stderr, "%u: %.*s: invalid rhs\n", line, len, start);
		return 1;
	}
	
	return 0;
}


typedef int (*check_syntax_t) (sm_str_P, sm_str_P);

static check_syntax_t check_syntax;

static struct syntax_tab {
	const char *name;
	check_syntax_t cfn;
} syntax_tab[] = {
	{ "access", check_access_syntax },
	{ "aliases", NULL },
	{ "qmgr_conf", NULL },
	{ NULL }
};
	  
static void
set_syntax(const char *arg)
{
	struct syntax_tab *sp;

	for (sp = syntax_tab; sp->name; sp++)
		if (strcmp(sp->name, arg) == 0) {
			check_syntax = sp->cfn;
			return;
		}
	fprintf(stderr, "unknown format: %s\n", arg);
	exit(EX_USAGE);
}
	
		   
static char buf[MAX_IN_LEN];
#if MAX_IN_LEN >= INT_MAX
ERROR _MAX_IN_LEN >= _INT_MAX
#endif

static int
sm_makemap(sm_map_P map, int sep, uint32_t flags)
{
	sm_ret_T ret;
	size_t len;
	sm_map_key_T key;
	sm_map_data_T data;
	char *delim, *rhs;
	int c;
	uint u;
	unsigned err_count = 0;
	
	line = 0;
	while (fgets(buf, sizeof(buf), stdin) != NULL) {
		++line;

		if (SM_IS_FLAG(flags, SMM_FL_SKIP_COMM)) {
			char *p;

			if (*buf == '\0' || *buf == SM_COMM_CHAR)
				continue;

			p = buf;
			while (*p != '\0' && ISSPACE(*p))
				p++;
			if (*p == '\0')
				continue;
		}

		/* find key/value separator */
		if (SM_IS_FLAG(flags, SMM_FL_WHSP_DELIM))
			delim = strpbrk(buf, " \t");
		else
			delim = strchr(buf, sep);

		if (NULL == delim) {
			/* complain about missing delim */
			fprintf(stderr,
				"%u: no key/value separator found\n",
				line);
			continue;
		}

		SM_ASSERT(delim < buf + sizeof(buf));
		SM_ASSERT(delim >= buf);
		u = (delim - buf) + 1;
		*delim = '\0';
		c = '\0';

		/* skip over leading space; make this optional? */
		if (SM_IS_FLAG(flags, SMM_FL_SKIP_LEAD_SP)) {
			while (u < sizeof(buf) && (c = buf[u]) != '\0'
			       && isspace(c) && c != '\n')
				++u;
		}
		if (u >= sizeof(buf) || c == '\0' || c == '\n') {
			fprintf(stderr, "%u: rhs is empty, u=%u\n",
				line, u);
			continue;
		}
		rhs = buf + u;
		while (u < sizeof(buf) && (c = buf[u]) != '\0' && c != '\n')
			++u;
		if (u >= sizeof(buf) || c != '\n') {
			fprintf(stderr, "%u: can't find end of rhs\n",
				line);
			continue;
		}
		buf[u] = '\0';

		len = strlen(buf);
		if (SM_IS_FLAG(flags, SMM_FL_LWR_KEY)) {
			uint j;

			for (j = 0; j < len; j++) {
				c = buf[j];
				if (ISUPPER(c))
					buf[j] = TOLOWER(c);
			}
		}

		if (!SM_IS_FLAG(flags, SMM_FL_EMPTY_KEY) && len == 0) {
			fprintf(stderr,"%u: empty key\n", line);
			continue;
		}
		sm_str_assign(key, NULL, (uchar *)buf, len, len);
		sm_str_assign(data, NULL, (uchar *)rhs, strlen(rhs),
				strlen(rhs));
		if (Verbose > 1) {
			fprintf(stderr, "lhs='%s' [%d], rhs='%s' [%d]\n"
				, (char *)sm_str_data(&key)
				, sm_str_getlen(&key)
				, (char *)sm_str_data(&data)
				, sm_str_getlen(&data)
				);
		}
		/* --gray: */
		if (check_syntax && check_syntax(&key, &data))
			err_count++;

		ret = sm_map_add(map, &key, &data, SMMAP_AFL_UNIQUE);
		if (sm_is_err(ret)) {
			fprintf(stderr, "%u: sm_map_add=failed, error=%s\n",
				line, smerr2txt(ret));
			sm_map_close(map, 0);
			return ret;
		}
	}

	ret = sm_map_close(map, 0);
	/* --gray */
	if (err_count && SM_IS_FLAG(flags, SMM_FL_STRICT_SYNTAX_CHECK))
		ret = sm_err_temp(EINVAL);
	return ret;
}

static int
sm_openmap(const char *mapname, const char *mapfile, const char *maptype, int sep, uint32_t flags)
{
	sm_ret_T ret;
	sm_maps_P maps;
	sm_map_P map;
	sm_cstr_P mtype, mname;

	maps = NULL;
	mtype = mname = NULL;

	ret = sm_maps_init(&maps);
	if (NULL == maps)
		return ret;

	mtype = sm_cstr_scpyn0((const uchar *)maptype, strlen(maptype));
	if (NULL == mtype) {
		ret = sm_err_temp(ENOMEM);
		goto error;
	}

	mname = sm_cstr_scpyn0((const uchar *)mapname, strlen(mapname));
	if (NULL == mname) {
		ret = sm_err_temp(ENOMEM);
		goto error;
	}

	ret = sm_bdb_class_create(maps);
	if (sm_is_err(ret))
		goto error;
#if MTA_USE_TINYCDB
	ret = sm_cdb_class_create(maps);
	if (sm_is_err(ret))
		goto error;
#endif

#if 0
	/* "delete" the file to make sure the result won't contain garbage? */
	ret = truncate(mapfile, 0);
	if (sm_is_err(ret))
		goto error;
#endif /* 0 */
	map = NULL;
	ret = sm_map_open(maps, mname, mtype, SMAP_MODE_CREATE, mapfile,
			SMAP_MODE_RDWR, &map, SMPO_END);
	if (sm_is_err(ret))
		goto error;

	ret = sm_makemap(map, sep, flags);
	if (sm_is_err(ret))
		goto error;

	ret = sm_maps_term(maps);
	SM_CSTR_FREE(mtype);
	SM_CSTR_FREE(mname);
	return ret;

  error:
	(void) sm_maps_term(maps);
	return ret;
}

static void
usage(const char *prg)
{
	fprintf(stderr, "usage: %s [options]\n"
		"-e        exit with error on syntax errors (see -H)\n"
		"-E        allow empty key\n"
		"-f        do not convert keys to lower case\n"
		"-F file   name of db file [%s]\n"
		"-H format validate input syntax according to format\n"
		"          valid formats are: access, aliases, qmgr_conf\n"
		"-l n      list available map classes (n: verbosity)\n"
		"-n name   name of map [%s]\n"
		"-s        do not ignore comment or empty lines\n"
		"-T type   map type [%s]\n"
		"-t c      separator [%c]\n"
		"-u        unlink db file in case of errors\n"
		"-V        increase verbosity\n"
		"-w        use whitespace as separator\n"
		"Options -t and -w are mutually exclusive.\n"
		"\n%s creates a %s map from the data provided via stdin.\n"
		"Entries must be of the form\n"
		"lhs%crhs\n"
		"(where '%c' is the default separator)\n"
		"Entries cannot span multiple lines.\n"
		, prg
		, MAPC_FILE
		, MAPC_NAME
		, MAPC_TYPE
		, SEP
		, prg
		, MAPC_TYPE
		, SEP
		, SEP
		);
	exit(EX_USAGE);
}

int
main(int argc, char *argv[])
{
	int c, sep;
	uint32_t flags;
	int list_mapc;
	bool rmonfail;
	char *mapname, *mapfile, *maptype, *prg;

	prg = argv[0];
	sep = (int) SEP;
	mapname = MAPC_NAME;
	mapfile = MAPC_FILE;
	maptype = MAPC_TYPE;
	list_mapc = -1;
	flags = SMM_FL_LWR_KEY|SMM_FL_SKIP_LEAD_SP|SMM_FL_SKIP_COMM;
	rmonfail = false;
#if !SM_ALLOW_ROOT
	if (getuid() == 0 || geteuid() == 0) {
		fprintf(stderr,
			"%s: ERROR: do not run this as super-user!\n",
			prg);
		exit(EX_USAGE);
	}
#endif
	while ((c = getopt(argc, argv, "eEfF:hH:l:n:SsT:t:uVw")) != -1) {
		switch (c) {
		  case 'e':
			flags |= SMM_FL_STRICT_SYNTAX_CHECK;
			break;
		  case 'E':
			flags |= SMM_FL_EMPTY_KEY;
			break;
		  case 'f':
			flags &= ~SMM_FL_LWR_KEY;
			break;
		  case 'F':
			SM_STRDUP_OPT(mapfile, optarg);
			break;
		  case 'H':
			set_syntax(optarg);
			break;
		  case 'l':
			list_mapc = (int) strtol(optarg, NULL, 0);
			break;
		  case 'n':
			SM_STRDUP_OPT(mapname, optarg);
			break;
		  case 's':
			flags &= ~SMM_FL_SKIP_COMM;
			break;
		  case 'S':
			flags &= ~SMM_FL_SKIP_LEAD_SP;
			break;
		  case 'T':
			SM_STRDUP_OPT(maptype, optarg);
			break;
		  case 't':
			if (SM_IS_FLAG(flags, SMM_FL_WHSP_DELIM)) {
				usage(prg);
				/* NOTREACHED */
				return EX_USAGE;
			}
			sep = (int) (*optarg);
			break;
		  case 'u':
			rmonfail = true;
			break;
		  case 'V':
			++Verbose;
			break;
		  case 'w':
			if (sep != (int) SEP) {
				usage(prg);
				/* NOTREACHED */
				return EX_USAGE;
			}
			flags |= SMM_FL_WHSP_DELIM;
			break;
		  case 'h':
		  default:
			usage(prg);
			/* NOTREACHED */
			return EX_USAGE;
		}
	}
	if (list_mapc >= 0) {
		sm_ret_T ret;
		sm_maps_P maps;

		maps = NULL;
		ret = sm_maps_init(&maps);
		if (NULL == maps)
			exit(EX_OSERR);
		ret = sm_maps_create(maps);
		if (sm_is_err(ret))
			exit(EX_OSERR);
		c = sm_mapc_list(smioout, maps, list_mapc,
				SMMAPC_FL_FILE|SMMAPC_FL_ADD);
	}
	else {
		c = sm_openmap(mapname, mapfile, maptype, sep, flags);
		if (sm_is_err(c)) {
			fprintf(stderr, "%s: %s\n",
				prg, smerr2txt(c));
			if (rmonfail)
				unlink(mapfile);
		}
	}
	return c;
}
