/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: mtparse.c,v 1.6 2009/06/01 13:41:14 ca Exp $")
#include "sm/error.h"
#include "sm/io.h"
#include "smar.h"
#include "log.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"
#include "sm/smarcnf.h"
#include "sm/smarcnfdef.h"
#include "mtparse.h"

/*
**  SMAR_MT_PARSE -- Parse RHS of mt entry
**
**	Parameters:
**		smar_clt_ctx -- SMAR client context
**		mtrhs -- RHS from mailertable
**
**	Returns:
**		usual sm_error code
*/

/*
RHS      ::= "{" ["protocol" "=" "esmtp" ";"] [port] hosts "}" |
		     "{" "protocol" "=" "lmtp" ";" [port hosts] "}"
port     ::= "port" "=" int ";"
protocol ::= "protocol" "=" ( "lmtp" | "esmtp") ";"
hosts    ::= "host" "=" hostname  | ips
ips      ::= "ipv4" "=" "{" ipv4-list "}"
ipv4     ::= IPv4-address
*/

static sm_conf_definition_T const
ar_prot_names[] = {
{ SM_CONF_DEF_MAGIC, "esmtp",		sm_conf_type_choice_value,
	SMAR_PROT_ESMTP,
	0, NULL, 0, NULL, NULL, NULL,
	"ESMTP"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },
{ SM_CONF_DEF_MAGIC, "lmtp",	sm_conf_type_choice_value,
	SMAR_PROT_LMTP,
	0, NULL, 0, NULL, NULL, NULL,
	"LMTP"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },


/* Sentinel */
{ SM_CONF_DEF_MAGIC, NULL, 0, 0, 0, NULL, 0, NULL, NULL, NULL, NULL SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0)}
};


static sm_conf_definition_T
ar_mtrhs_defs[] = {
{ SM_CONF_DEF_MAGIC, "protocol",		sm_conf_type_choice,
	offsetof(ar_mtcnf_T, ar_mt_cnf_prot),	sizeof(uint),
	NULL, 0,
	ar_prot_names,	NULL,	NULL,
	"protocol name"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

{ SM_CONF_DEF_MAGIC, "port",		sm_conf_type_u32,
	offsetof(ar_mtcnf_T, ar_mt_cnf_port),	sizeof(ushort),
	NULL,
	0,	NULL,	NULL,	NULL,
	"port number"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

{ SM_CONF_DEF_MAGIC, "ipv4",		sm_conf_type_argv,
	offsetof(ar_mtcnf_T, ar_mt_cnf_ips),	SMAR_MAX_MT_IPV4,
	NULL,	0,	NULL,	NULL,	NULL,
	"List of hosts (IPv4 addresses)"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

{ SM_CONF_DEF_MAGIC, "host",		sm_conf_type_string,
	offsetof(ar_mtcnf_T, ar_mt_cnf_host),	0,
	NULL,	0,	NULL,	NULL,	NULL,
	"hostname"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

/* Sentinel */
{ SM_CONF_DEF_MAGIC, NULL, 0, 0, 0, NULL, 0, NULL, NULL, NULL, NULL SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0)}
};

sm_ret_T
smar_mt_parse(smar_clt_ctx_P smar_clt_ctx, sm_str_P mtrhs, ar_mtcnf_P ar_mtcnf)
{
	int err;
	uint ui;
	char *s;
#define AR_MT_RHS "mtrhs"

	SM_IS_SMAR_CLT_CTX(smar_clt_ctx);
	SM_IS_BUF(mtrhs);
	SM_REQUIRE(ar_mtcnf != NULL);

	smar_clt_ctx->smac_cnf = sm_conf_new(AR_MT_RHS);
	if (NULL == smar_clt_ctx->smac_cnf) {
		err = errno;

		sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_mt_parse, sm_conf_new=NULL, errno=%d\n",
			err);
		return sm_error_temp(SM_EM_AR, ENOMEM);
	}

	err = sm_conf_read_data(smar_clt_ctx->smac_cnf,
				(char *)sm_str_getdata(mtrhs),
				sm_str_getlen(mtrhs), true);
	if (err != 0) {
		sm_prt_conferr(AR_MT_RHS, smar_clt_ctx->smac_cnf, err, smioerr);
		sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_mt_parse, sm_conf_read_data=%m\n",
			err);
		sm_conf_destroy(smar_clt_ctx->smac_cnf);
		smar_clt_ctx->smac_cnf = NULL;
		return err;
	}

	err = sm_conf_scan(smar_clt_ctx->smac_cnf, ar_mtrhs_defs, 0, ar_mtcnf);
	if (err != 0) {
		sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_mt_parse, sm_conf_scan=%m\n",
			err);
		goto error;
	}

	/*
	**  HACK: "parse" configuration data.
	**  This will go away as soon as the configuration
	**  definition evolves... (array of IPv4 addresses)
	*/

	while ((ui = ar_mtcnf->ar_mt_cnf_n_ipv4s) < SMAR_MAX_MT_IPV4
		&& (s = ar_mtcnf->ar_mt_cnf_ips[ui]) != NULL)
	{
		err = sm_inet_a2ipv4(s, NULL, &ar_mtcnf->ar_mt_cnf_ipv4s[ui]);
		if (sm_is_err(err)) {
			sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
				AR_LCAT_CONFIG, AR_LMOD_CONFIG,
				SM_LOG_ERR, 8,
				"sev=ERROR, func=smar_mt_parse, mt_ipv4=%s, status=not_an_IPv4_address\n", s);
			goto error;
		}
		++ar_mtcnf->ar_mt_cnf_n_ipv4s;
	}

	if (ar_mtcnf->ar_mt_cnf_n_ipv4s > 0 && ar_mtcnf->ar_mt_cnf_host != NULL) {
		sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_mt_parse, status=IPv4_address_and_hostname_defined\n");
			err = sm_err_perm(EINVAL);
	}

	if (0 == ar_mtcnf->ar_mt_cnf_port) {
		if (SMAR_PROT_LMTP == ar_mtcnf->ar_mt_cnf_prot &&
		    (ar_mtcnf->ar_mt_cnf_n_ipv4s > 0 ||
			 ar_mtcnf->ar_mt_cnf_host != NULL))
		{
			sm_log_write(smar_clt_ctx->smac_ar_ctx->smar_lctx,
				AR_LCAT_CONFIG, AR_LMOD_CONFIG,
				SM_LOG_ERR, 8,
				"sev=ERROR, func=smar_mt_parse, protocol=lmtp, status=no_port_defined\n");
			err = sm_err_perm(EINVAL);
		}
#if 0
		else
			ar_mtcnf->ar_mt_cnf_port = SMTPC_PORT;
#endif
	}

/*
	sm_conf_prt_conf(ar_mtrhs_defs, smar_clt_ctx->smac_cnf, smioout);
*/

  error:
	return err;
}
